/*
 * Author: Konrad Rzeszutek <conradus@mindless.com>
 * Date:   Oct 19th 1999
 * File:   Editor3D/FileObject.java
 *
 * Notes:
 *
 *   This nice little class reads in the all the information
 * from the specified file. The queuries return the info in a nice
 * form usuable by other objects.
 *
 */
 package Editor3D;
 
import java.io.*;
import java.util.*;

public class FileObject {

   /*
    * Initializes the FileObject class and reads in the 
    * 'filename' file -- this "populates" the FaceList and WC system.
    *
    * require:
    *   filename instanceof String.
    */
    public FileObject(String filename) {
 	
	this.populate(filename);
    }
    
    /*
     * Read in one token.
     *
     * return:
     *  true if there is an error.
     */
    private boolean read() {
    
	boolean err = false;
	
	try {
	    err = st.nextToken() == StreamTokenizer.TT_EOF;
	    }
	    catch ( IOException e) {
		 System.out.println("Can't read the file!");
	    }
	return err;
    }
    
    /*
     * Simple method reading in four consecutive double's.
     *
     * require:
     *  _must_ be called from populate().
     *
     */
    private double[] readThreeDoubles() {
    
	double aData[] = new double[3];
	boolean eof = false;
	//read(); // Flush
	for (int i = 0; (i < 3) && (!eof); i++) {
	    eof = read();
	    aData[i] = st.nval;
	    //	    System.out.print("["+st.nval+"]");
	} 
	//	System.out.println(".");
	return aData;
    }
    private double[] readFourDoubles() {
    
	double aData[] = new double[4];
	boolean eof = false;
	//read(); // Flush
	for (int i = 0; (i < 4) && (!eof); i++) {
	    eof = read();
	    aData[i] = st.nval;
	    //System.out.print("["+st.nval+"]");
	} 
	//System.out.println(".");
	return aData;
    }

        
    /*
     * Read in the contenst of the 'f' file. Catagorize it
     * accordingly to the file specification.
     *
     * require:
     *  f instanceof String.
     */
    private void populate(String f) {

	boolean eof = false;
	double[] aHithYon;   // array for Hith and yon
	double[] aLight;     // array for light src
	double[] aVec;	     // Array for one vector point
	int iListVectLng;    // lenght of the VectorObjects list
	int iListFaceLng;    // length of the list of faces.
	int iFaces;          // amount of faces
	
	try {
	    // Open the file.
	    st = new StreamTokenizer(new BufferedReader(new FileReader(f)));
	    st.eolIsSignificant(true);
	    st.parseNumbers();

	} catch (Exception e) {
	   System.out.println("Can't open the @#^&#%#@ file");
	}
	

	aVec = new double[3];
	aLight = new double[4];
	aHithYon = new double[2];
	
	read();
	iListVectLng = (int)st.nval; // Find out how many total points we will have
	read();
	iListFaceLng = (int)st.nval; // Find out how many total faces we will have
	
	listVectObj = new Vector (iListVectLng);
	listFaceObj = new Vector (iListFaceLng);

			
	// First read in the points. There should be only iListVecLng of them
	    
	for (int i = 0; (i < iListVectLng) && (!eof); i++) {
	    read(); // flush
	    for (int j = 0; (j < 3) && (!eof); j++) {// We only have X,Y,Z co-ordinate system, hence 3
		    eof = read();
		    aVec[j] = st.nval;
		}
		// Add a new VectorObject the list.
		listVectObj.addElement( (Object) new VectorObject(
		    aVec[VectorObject.X_POS],
		    aVec[VectorObject.Y_POS],
		    aVec[VectorObject.Z_POS]));
	    }
	    
	// Read in the Faces.

	for (int i = 0; (i < iListFaceLng) && (!eof);i++) {
	    read(); // flush
	    // Find out how many points in this face.
	    eof = read();
	    iFaces = (int) st.nval;
    	    listFaceObj.addElement( (Object) new FaceObject(iFaces));
	    // Read in the references to the points.
	    for (int j = 0; (j < iFaces) && (!eof); j++) {
		eof = read();
		// Add the required VectorObject to the FaceObject.
		((FaceObject) (listFaceObj.lastElement())).add( (VectorObject)listVectObj.elementAt((int)st.nval-1));
	    }

	}
	
	// Initalize the ViewPort -- it has all the VRP, VPN, .. , window info.
	viewPort = new ViewPort(); // Make the viewport.	
	read(); // flush
	// Read in the VRP
	aVec = readThreeDoubles();
	viewPort.setVRP(new VectorObject(aVec[VectorObject.X_POS], aVec[VectorObject.Y_POS], aVec[VectorObject.Z_POS]));
	// Read in VPN
	aVec = readThreeDoubles();
	viewPort.setVPN(new VectorObject(aVec[VectorObject.X_POS], aVec[VectorObject.Y_POS], aVec[VectorObject.Z_POS]));
	// Read in VUP
	aVec = readThreeDoubles();
	viewPort.setVUP(new VectorObject(aVec[VectorObject.X_POS], aVec[VectorObject.Y_POS], aVec[VectorObject.Z_POS]));
	// Read in COP
	aVec = readThreeDoubles();
	viewPort.setCOP(new VectorObject(aVec[VectorObject.X_POS], aVec[VectorObject.Y_POS], aVec[VectorObject.Z_POS]));
	// Read in hith and yon
   	for (int i = 0; (i < 2) && (!eof); i++) {
	    eof = read();
	    aHithYon[i] = st.nval;
	} 
	// Read in the Window info
	viewPort.setUV(readFourDoubles());
	// And that's it. We won't care about the lighting (yet)
	//aLight = readFourDoubles();
	//viewPort.setLight(aLight);
    }
    
    
    /*
     * Returns the ViewPort read in from the file.
     *
     * return:
     *  ViewPort || null
     *
     */
    public ViewPort getViewPort() {
    
	return viewPort;
    }
    
    /*
     * Returns the list of faces read in from the file.
     *
     * return:
     *  null || length() >= 1
     */     
    public Vector getFaceList() {
    
	return listFaceObj;
	
    }

    /*
     * Returns a list of all the vectors read in.
     *
     * return
     * getVectorList.length() >= 0
     *
     */
    public Vector getVectorList() {

	return listVectObj;
    }
    private Vector listFaceObj;    // List of faces.
    private Vector listVectObj;    // List of 3D points.
    private ViewPort viewPort;     // Our UV viewpoint
    private StreamTokenizer st;    // To read data in.
}
